/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_COLLISION_AVOIDER_H
#define HPP_LOS_COLLISION_AVOIDER_H

#include "hpplocalstepper/tasks/hppLosTask.hh"
#include "hpphik/hppHikSlackSolver.hh"




class LosRobotCylinder
{
    public:
        ChppLosRobotLink* link;
        CjrlJoint *joint;
        vector3d A,B,AB;
        vector3d locAB,locA,locB;
        double normAB;
        double radius;
};


class LosRobotSphere
{
    public:
        ChppLosRobotLink* link;
        CjrlJoint *joint;
        vector3d S, locS;
        double radius;
};


class LosExtPlane
{
    public:
        vector3d P, N; // point and normal in world frame
        double ds;
};

class LosExtCylinder
{
    public:
        vector3d A,B,AB;
        double normAB;
        double radius;
};

class LosLinkFootprint
{
    public:
        ChppLosLink* link;
        vector3d P1,P2,P3,P4;
        double ds;
};

class LinkDamper
{
    public:
        ChppLosLink* link;
        vector3d absPoint1,absPoint2;
        vector3d ntojoint;
        double distance1,distance2;
        double Di,Ds,Di_Ds;
};

class ExtDamper
{
    public:
        ChppLosRobotLink* link;
        CjrlJoint* joint;
        vector3d locPoint, absPoint;
        vector3d ntojoint;
        double distance;
        double Di,Ds,Di_Ds;
};

class SelfDamper
{
    public:
        ChppLosRobotLink* link;
        CjrlJoint* joint1;
        CjrlJoint* joint2;
        vector3d locPoint1;
        vector3d locPoint2;
        vector3d n12;
        double distance;
        double Di,Ds,Di_Ds;
};



/**
A task for HRP2 to avoid having the arms collide with the trunc or the thighs and both hand from approching the head
\ingroup tasks
 */
class ChppLosCollisionAvoider: public ChppLosTask
{
    public:
        /**
         \brief Constructor
         */
        ChppLosCollisionAvoider ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inLink );
        /**
        \brief Set the area of translation brake inside the authorized configurations region boudaries
         */
        void setMargins ( double inDi, double inDs, double inXi );
        /**
        \brief Add a sphere
         */
        unsigned int addRobotSphere ( LosRobotSphere* inS );
        /**
        \brief Add a sphere
         */
        unsigned int addRobotCylinder ( LosRobotCylinder* inC );
        /**
        \brief Add a sphere
         */
        unsigned int addLinkFootprint ( LosLinkFootprint* inC );
        /**
        \brief Add an exterior cylinder
         */
        unsigned int addExtCylinder ( LosExtCylinder* inC );
        /**
        \brief Add a plane
         */
        unsigned int addExtPlane ( LosExtPlane* inP );
        /**
        \brief Add a pair of cylinders for collision check
         */
        bool makeRobotCylinderPair ( unsigned int id_1, unsigned int id_2 );
        /**
        \brief Add a pair of cylinder and extCylinder for collision check
         */
        bool makeRobotCylinderExtPair ( unsigned int id_1, unsigned int id_2 );
        /**
        \brief Add a pair of cylinder-plane for collision check
         */
        bool makeRobotCylinderPlanePair ( unsigned int id_1, unsigned int id_2 );
        /**
        \brief Add a pair of cylinder-plane for collision check
         */
        bool makeLinkFootprintPlanePair ( unsigned int id_1, unsigned int id_2 );
        /**
        \brief Add a pair of spheres for collision check
         */
        bool makeRobotCylinderSpherePair ( unsigned int id_1, unsigned int id_2 );
        /**
        \brief Add a pair of cylinder and sphere for collision check
         */
        bool makeRobotSpherePair ( unsigned int id_1, unsigned int id_2 );
        /**
        \brief Inherited.
         */
        bool isEqualityTask() const {return false;};
        /**
        \brief Inherited.
         */
        void computeConstraint();
        /**
        \brief Inherited.
         */
        const vectorN& currentValue();

        /**
        \brief Destructor
         */
        ~ChppLosCollisionAvoider();

    private:
        void updateModel();

        std::vector<LosRobotSphere*> attRobotSpheres;
        std::vector<LosRobotCylinder*> attRobotCylinders;
        std::vector<LosExtCylinder*> attExtCylinders;
        std::vector<LosLinkFootprint*> attLinkFootprints;
        std::vector<LosExtPlane*> attExtPlanes;

        std::vector< std::pair<LosLinkFootprint*, LosExtPlane*> > attLinkFootprintPlanePairs;
        std::vector< std::pair<LosRobotCylinder*, LosExtPlane*> > attRobotCylinderPlanePairs;
        std::vector< std::pair<LosRobotCylinder*, LosExtCylinder*> > attRobotCylinderExtPairs;
        std::vector< std::pair<LosRobotCylinder*, LosRobotCylinder*> > attRobotCylinderPairs;
        std::vector< std::pair<LosRobotCylinder*, LosRobotSphere*> > attRobotCylinderSpherePairs;
        std::vector< std::pair<LosRobotSphere*, LosRobotSphere*> > attRobotSpherePairs;

        std::vector<ExtDamper> attRobotCylinderPlaneDampers;
        std::vector<LinkDamper> attLinkFootprintPlaneDampers;
        std::vector<ExtDamper> attRobotCylinderExtDampers;
        std::vector<SelfDamper> attRobotCylinderDampers,attRobotSphereDampers,attRobotCylinderSphereDampers;
        ChppHikSlackSolver* attSolver;
        ChppHikLinearSystem* attSystem;

        void calcLinkConstraint ( LinkDamper& outDamper, unsigned int constraintRank );
        void calcExtConstraint ( ExtDamper& outDamper, unsigned int constraintRank );
        void calcConstraint ( SelfDamper& outDamper, unsigned int constraintRank );

        bool checkLinkFootprintPlanePair ( unsigned int pairRank );
        bool checkRobotSpherePair ( unsigned int pairRank );
        bool checkRobotCylinderPlanePair ( unsigned int pairRank );
        bool checkRobotCylinderSpherePair ( unsigned int pairRank );
        bool checkRobotCylinderPair ( unsigned int pairRank );
        bool checkRobotCylinderExtPair ( unsigned int pairRank );

        void crossProd3 ( const vector3d& v1, const vector3d& v2, vector3d& retV );

        double attXi,attDi,attDs;

        matrixNxP tempJac3d,tempJac3dLong,noutM,jac1d;


};

#endif
