/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_TASK_H
#define HPP_LOS_TASK_H

#include "hpphik/hppHikLinearSystem.hh"
#include "hpplocalstepper/hppLosCentipede.hh"

/**
Abstract interface for a task. A task produces a linear system that can be composed either of equalities or inequalities. The Linear system is dependant on the state of the ChppLosCentipede object passed to the task's constructor. The linear system is updated through call of method computeConstraint() and can be retrieved through call of method constraint().
\ingroup tasks
*/
class ChppLosTask
{
    public:
        /**
        \brief Constructor
        */
        ChppLosTask(ChppLosCentipede& inCentipede)
        {
            attCentipede = &inCentipede;
            attConstraint = new ChppHikLinearSystem( attCentipede->numberDof() );
            attDamp = 0.0;
            attEndTime = 0.0;
            attStartTime = 0.0;
        }
        /**
        \brief Destructor
        */
        virtual ~ChppLosTask(){delete attConstraint;}
        /**
        \brief Compute the linear system to be solved for this task at the current state
         */
        virtual void computeConstraint() = 0;
        /**
        \brief Return the value of the task. The task is strictly solved if the norm of the returned vector is null.
         */
        virtual const vectorN& currentValue() = 0;
        /**
        \brief Tell if this task is an equality task
         */
        virtual bool isEqualityTask() const = 0;
        /**
        \brief Get the linear constraint to be solved for this task at the current state
         */
        virtual ChppHikLinearSystem& constraint() {return *attConstraint;};
        /**
        \brief Get the damping factor associated with the linear system
         */
        virtual double dampingFactor() {return attDamp;};
        /**
        \brief Set the damping factor associated with the linear system
         */
        virtual void dampingFactor ( double inF ) {attDamp = inF;};
        /**
        \brief tag with a start time
        */
        void startTime(double inT){attStartTime = inT;};
        /**
        \brief get the start time (return 0 if not set)
         */
        double startTime(){return attStartTime;};
        /**
        \brief tag with an end time
         */
        void endTime(double inT){attEndTime = inT;};
        /**
        \brief get the end time (return 0 if not set)
         */
        double endTime(){return attEndTime;};

    protected:
        virtual bool resize()
        {
            if (attConstraint->matrix().size2() != attCentipede->numberDof())
            {
                attConstraint->matrix().resize(attConstraint->matrix().size1(),attCentipede->numberDof());
                attConstraint->matrix().clear();
                return true;
            }
            return false;
        };
        ChppLosCentipede* attCentipede;
        ChppHikLinearSystem* attConstraint;
        double attDamp,attStartTime,attEndTime;
        vectorN attValue;
};

/**
A task that uses a joint in the robot link
\ingroup tasks
*/
class ChppLosTaskOnJoint: public ChppLosTask
{
    public:
        ChppLosTaskOnJoint(ChppLosCentipede& inCentipede,CjrlJoint& inJoint):ChppLosTask(inCentipede)
        {
            attJoint = &inJoint;
        }
        /**
        \brief Get reference joint
         */
        CjrlJoint* joint()
        {
            return attJoint;
        }
        /**
        \brief Get reference joint
         */
        void joint(CjrlJoint* inJoint)
        {
            attJoint = inJoint;
        }
        
    protected:
        CjrlJoint* attJoint;
};

#endif
