/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpplocalstepper/hppLosLink.hh"
#include "hppGikTools.h"

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J
using namespace boost::numeric::ublas;

ChppLosLink::ChppLosLink ( unsigned int inNumberDof, bool inRootIsRight, const ChppLosFootprintArea& leftStepArea,  const ChppLosFootprintArea& righttStepArea, const ChppLosFootprintShape& inLeftFootShape, const ChppLosFootprintShape& inRightFootShape )
{
    attRootIsRight = inRootIsRight;
    attRank = 0;
    attNumberDof = inNumberDof;

    attLeftArea = leftStepArea;
    attRightArea = righttStepArea;
    attLeftShape = inLeftFootShape;
    attRightShape = inRightFootShape;

    attJP.jacLYawinR.resize ( attNumberDof );
    attJP.jacLinR.resize ( 2,attNumberDof );
    attJP.jacLUinR.resize ( 2,attNumberDof );
    attJP.jacLDinR.resize ( 2,attNumberDof );
    attJP.jacRYawinL.resize ( attNumberDof );
    attJP.jacRinL.resize ( 2,attNumberDof );
    attJP.jacRUinL.resize ( 2,attNumberDof );
    attJP.jacRDinL.resize ( 2,attNumberDof );

    if ( inRootIsRight )
    {
        attRootprint = &attRightprint;
        attFrontprint = &attLeftprint;
    }
    else
    {
        attRootprint = &attLeftprint;
        attFrontprint = &attRightprint;
    }

    attParent = 0;

    attConfiguration = zero_vector<double> ( inNumberDof );
    attUb = scalar_vector<double> ( inNumberDof,1e10 );
    attLb = scalar_vector<double> ( inNumberDof,-1e10 );
}

bool ChppLosLink::rootIsRight() const
{
    return attRootIsRight;
}

bool ChppLosLink::getUpperVelocityBounds(vectorN& outVec)
{
    if (outVec.size() < attRank+attNumberDof)
    {
        std::cout << "ChppLosLink::getUpperVelocityBounds bad vector size" << std::endl;
        return false;
    }
    noalias(subrange(outVec,attRank,attRank+attNumberDof)) = attUb;
    return true;
}

bool ChppLosLink::getLowerVelocityBounds(vectorN& outVec)
{
    if (outVec.size() < attRank+attNumberDof)
    {
        std::cout << "ChppLosLink::getLowerVelocityBounds bad vector size" << std::endl;
        return false;
    }
    noalias(subrange(outVec,attRank,attRank+attNumberDof)) = attLb;
    return true;
}

bool ChppLosLink::getConfiguration(vectorN& outVec)
{
    if (outVec.size() < attRank+attNumberDof)
    {
        std::cout << "ChppLosLink::getLowerVelocityBounds bad vector size" << std::endl;
        return false;
    }
    noalias(subrange(outVec,attRank,attRank+attNumberDof)) = attConfiguration;
    return true;
}

void ChppLosLink::computeRelativeCoordinates()
{
    {
        double crth = cos ( attRightprint.th );
        double srth = sin ( attRightprint.th );
        double dx = attLeftprint.x - attRightprint.x;
        double dy = attLeftprint.y - attRightprint.y;

        attJP.LinRx = crth *dx + srth *dy;
        attJP.LinRy = - srth *dx + crth *dy;
        attJP.LinRth = attLeftprint.th-attRightprint.th;

        while (attJP.LinRth > M_PI)
            attJP.LinRth -= 2*M_PI;
        while (attJP.LinRth < -M_PI)
            attJP.LinRth += 2*M_PI;
        
        double cdth = cos ( attJP.LinRth );
        double sdth = sin ( attJP.LinRth );

        attJP.ULinRx = attJP.LinRx + attLeftShape.xUpRight*cdth - sdth * attLeftShape.yUpRight;
        attJP.ULinRy = attJP.LinRy + attLeftShape.xUpRight*sdth + cdth * attLeftShape.yUpRight;
        attJP.DLinRx = attJP.LinRx + attLeftShape.xDownRight*cdth - sdth * attLeftShape.yDownRight;
        attJP.DLinRy = attJP.LinRy + attLeftShape.xDownRight*sdth + cdth * attLeftShape.yDownRight;
    }

    {
        double clth = cos ( attLeftprint.th );
        double slth = sin ( attLeftprint.th );
        double dx = attRightprint.x - attLeftprint.x;
        double dy = attRightprint.y - attLeftprint.y;

        attJP.RinLx = clth *dx + slth *dy;
        attJP.RinLy = - slth *dx + clth *dy;
        attJP.RinLth = attRightprint.th-attLeftprint.th;
        
        while (attJP.RinLth > M_PI)
            attJP.RinLth -= 2*M_PI;
        while (attJP.RinLth < -M_PI)
            attJP.RinLth += 2*M_PI;

        double cdth = cos ( attJP.RinLth );
        double sdth = sin ( attJP.RinLth );


        attJP.URinLx = attJP.RinLx + attRightShape.xUpLeft*cdth - sdth * attRightShape.yUpLeft;
        attJP.URinLy = attJP.RinLy + attRightShape.xUpLeft*sdth + cdth * attRightShape.yUpLeft;
        attJP.DRinLx = attJP.RinLx + attRightShape.xDownLeft*cdth - sdth * attRightShape.yDownLeft;
        attJP.DRinLy = attJP.RinLy + attRightShape.xDownLeft*sdth + cdth * attRightShape.yDownLeft;
    }
    
   
}

void ChppLosLink::parent ( ChppLosLink* inP )
{
    attParent = inP;
}

ChppLosLink* ChppLosLink::parent() const
{
    return attParent;
}

const ChppLosFootprint& ChppLosLink::frontprint() const
{
    return *attFrontprint;
}
const ChppLosFootprint& ChppLosLink::rootprint() const
{
    return *attRootprint;
}
bool ChppLosLink::isRootprintRight() const
{
    return attRootIsRight;
}

void ChppLosLink::rank ( unsigned int inRank )
{
    attRank = inRank;
}
unsigned int ChppLosLink::rank() const
{
    return attRank;
}

unsigned int ChppLosLink::numberDof() const
{
    return attNumberDof;
}

const ChppLosFootprint& ChppLosLink::leftprint() const
{
    return attLeftprint;
}

const ChppLosFootprint& ChppLosLink::rightprint() const
{
    return attRightprint;
}


ChppLosSupportPolygonLink::ChppLosSupportPolygonLink ( bool rootIsRight, double inDX, double inDY, double inDth , const ChppLosFootprintArea& leftStepArea,  const ChppLosFootprintArea& righttStepArea, const ChppLosFootprintShape& inLeftFootShape, const ChppLosFootprintShape& inRightFootShape ) :ChppLosLink ( 3,rootIsRight,leftStepArea,righttStepArea,inLeftFootShape,inRightFootShape )
{
    attConfiguration ( 0 ) = inDX;
    attConfiguration ( 1 ) = inDY;
    attConfiguration ( 2 ) = inDth;
    
    while (attConfiguration ( 2 ) > M_PI)
        attConfiguration ( 2 ) -= 2*M_PI;
    while (attConfiguration ( 2 ) < -M_PI)
        attConfiguration ( 2 ) += 2*M_PI;
}

bool ChppLosSupportPolygonLink::getJacobian ( const vector3d& inPoint, matrixNxP& outJacobian, unsigned int inStart )
{
    if ( ( outJacobian.size2() < inStart + 3 ) || ( outJacobian.size1() !=6 ) )
    {
        std::cout << "ChppLosSupportPolygonLink::getJacobian -Error bad size for outJacobian " << std::endl;
        return false;
    }

    double dx = inPoint[0] - attFrontprint->x;
    double dy = attFrontprint->y - inPoint[1];
    unsigned int rr = inStart;

    outJacobian ( 0,rr ) = attRootprint->UXx;
    outJacobian ( 1,rr ) = attRootprint->UXy;
    rr++;
    outJacobian ( 0,rr ) = attRootprint->UYx;
    outJacobian ( 1,rr ) = attRootprint->UYy;
    rr++;
    outJacobian ( 0,rr ) = dy;
    outJacobian ( 1,rr ) = dx;
    outJacobian ( 5,rr ) = 1;

    return true;
}

bool ChppLosSupportPolygonLink::getPositionJacobian ( const vector3d& inPoint, matrixNxP& outJacobian, unsigned int inStart )
{
    if ( ( outJacobian.size2() < inStart + 3 ) || ( outJacobian.size1() !=3 ) )
    {
        std::cout << "ChppLosSupportPolygonLink::getPositionJacobian -Error bad size for outJacobian " << std::endl;
        return false;
    }

    double dx = inPoint[0] - attFrontprint->x;
    double dy = attFrontprint->y - inPoint[1];
    unsigned int rr = inStart;

    outJacobian ( 0,rr ) = attRootprint->UXx;
    outJacobian ( 1,rr ) = attRootprint->UXy;
    rr++;
    outJacobian ( 0,rr ) = attRootprint->UYx;
    outJacobian ( 1,rr ) = attRootprint->UYy;
    rr++;
    outJacobian ( 0,rr ) = dy;
    outJacobian ( 1,rr ) = dx;


    return true;
}

bool ChppLosSupportPolygonLink::getOrientationJacobian ( matrixNxP& outJacobian, unsigned int inStart )
{
    if ( ( outJacobian.size2() < inStart + 3 ) || ( outJacobian.size1() !=3 ) )
    {
        std::cout << "ChppLosSupportPolygonLink::getOrientationJacobian -Error bad size for outJacobian " << std::endl;
        return false;
    }
    outJacobian ( 2,inStart+2 ) = 1;
    return true;
}

bool ChppLosSupportPolygonLink::applyConfiguration ( const ChppLosFootprint& prevleftprint, const ChppLosFootprint& prevrightprint, const vectorN& inConfiguration, bool computeRelativeFeetjacobians )
{
    if ( inConfiguration.size() < attRank + attNumberDof  )
    {
        std::cout << "ChppLosSupportPolygonLink::applyConfiguration -Error: bad vector size" << std::endl;
        return false;
    }

    //forward world positions and axes root foot
    if ( attRootIsRight )
        attRightprint = prevrightprint;
    else
        attLeftprint = prevleftprint;

    noalias(attConfiguration) = subrange(inConfiguration,attRank,attRank+attNumberDof);

    double cs = cos ( attRootprint->th );
    double ss = sin ( attRootprint->th );
    attFrontprint->x = attRootprint->x + cs*attConfiguration ( 0 ) - ss*attConfiguration ( 1 );
    attFrontprint->y = attRootprint->y + ss*attConfiguration ( 0 ) + cs*attConfiguration ( 1 );
    attFrontprint->th = attRootprint->th + attConfiguration ( 2 );
    
    while (attFrontprint->th > M_PI)
        attFrontprint->th -= 2*M_PI;
    while (attFrontprint->th < -M_PI)
        attFrontprint->th += 2*M_PI;

    attFrontprint->UXx = cos ( attFrontprint->th );
    attFrontprint->UXy = sin ( attFrontprint->th );
    attFrontprint->UYx = -attFrontprint->UXy;
    attFrontprint->UYy = attFrontprint->UXx;

    if (computeRelativeFeetjacobians)
        computeJacobianPack ();
    return true;
}

void ChppLosSupportPolygonLink::computeJacobianPack ( )
{
    //relative coordinates of left foot in right foot
    computeRelativeCoordinates();
    
    {
        attJP.jacLinR.clear();
        attJP.jacLYawinR.clear();
        if ( attRootIsRight )
        {
            attJP.jacLinR ( 0,0 ) = 1;
            attJP.jacLinR ( 1,1 ) = 1;

            attJP.jacLYawinR ( 2 ) = 1;

            attJP.jacLUinR = attJP.jacLinR;
            attJP.jacLUinR ( 0,2 ) = attJP.LinRy-attJP.ULinRy;
            attJP.jacLUinR ( 1,2 ) = attJP.ULinRx-attJP.LinRx;

            attJP.jacLDinR = attJP.jacLinR;
            attJP.jacLDinR ( 0,2 ) = attJP.LinRy-attJP.DLinRy;
            attJP.jacLDinR ( 1,2 ) = attJP.DLinRx-attJP.LinRx;
        }
        else
        {
            attJP.jacLinR ( 0,0 ) = -cos ( attConfiguration ( 2 ) );
            attJP.jacLinR ( 1,1 ) = attJP.jacLinR ( 0,0 );
            attJP.jacLinR ( 0,1 ) = -sin ( attConfiguration ( 2 ) );
            attJP.jacLinR ( 1,0 ) = -attJP.jacLinR ( 0,1 );
            double cs =cos ( attRightprint.th );
            double ss =sin ( attRightprint.th );
            double prlx = attLeftprint.x-attRightprint.x;
            double prly = attLeftprint.y-attRightprint.y;
            attJP.jacLinR ( 0,2 ) = -ss*prlx+cs*prly;
            attJP.jacLinR ( 1,2 ) = -cs*prlx-ss*prly;

            attJP.jacLYawinR ( 2 ) = -1;

            attJP.jacLUinR = attJP.jacLinR;
            attJP.jacLUinR ( 0,2 ) += attJP.ULinRy-attJP.LinRy;
            attJP.jacLUinR ( 1,2 ) += attJP.LinRx-attJP.ULinRx;

            attJP.jacLDinR = attJP.jacLinR;
            attJP.jacLDinR ( 0,2 ) += attJP.DLinRy-attJP.LinRy;
            attJP.jacLDinR ( 1,2 ) += attJP.LinRx-attJP.DLinRx;

        }
    }
    {
        attJP.jacRinL.clear();
        attJP.jacRYawinL.clear();
        if ( attRootIsRight )
        {
            attJP.jacRinL ( 0,0 ) = -cos ( attConfiguration ( 2 ) );
            attJP.jacRinL ( 1,1 ) = attJP.jacRinL ( 0,0 );
            attJP.jacRinL ( 0,1 ) = -sin ( attConfiguration ( 2 ) );
            attJP.jacRinL ( 1,0 ) = -attJP.jacRinL ( 0,1 );
            double cs =cos ( attLeftprint.th );
            double ss =sin ( attLeftprint.th );
            double plrx = attRightprint.x-attLeftprint.x;
            double plry = attRightprint.y-attLeftprint.y;
            attJP.jacRinL ( 0,2 ) = -ss*plrx+cs*plry;
            attJP.jacRinL ( 1,2 ) = -cs*plrx-ss*plry;

            attJP.jacRYawinL ( 2 ) = -1;

            attJP.jacRUinL = attJP.jacRinL;
            attJP.jacRUinL ( 0,2 ) += attJP.URinLy-attJP.RinLy;
            attJP.jacRUinL ( 1,2 ) += attJP.RinLx-attJP.URinLx;

            attJP.jacRDinL = attJP.jacRinL;
            attJP.jacRDinL ( 0,2 ) += attJP.DRinLy-attJP.RinLy;
            attJP.jacRDinL ( 1,2 ) += attJP.RinLx-attJP.DRinLx;
        }
        else
        {
            attJP.jacRinL ( 0,0 ) = 1;
            attJP.jacRinL ( 1,1 ) = 1;

            attJP.jacRYawinL ( 2 ) = 1;

            attJP.jacRUinL = attJP.jacRinL;
            attJP.jacRUinL ( 0,2 ) = attJP.RinLy-attJP.URinLy;
            attJP.jacRUinL ( 1,2 ) = attJP.URinLx-attJP.RinLx;

            attJP.jacRDinL = attJP.jacRinL;
            attJP.jacRDinL ( 0,2 ) = attJP.RinLy-attJP.DRinLy;
            attJP.jacRDinL ( 1,2 ) = attJP.DRinLx-attJP.RinLx;
        }
    }
}

ChppLosRobotLink::ChppLosRobotLink ( ChppGikStandingRobot& inRobot, bool rootIsRight, const ChppLosFootprintArea& leftStepArea,  const ChppLosFootprintArea& righttStepArea, const ChppLosFootprintShape& inLeftFootShape, const ChppLosFootprintShape& inRightFootShape ) :ChppLosLink ( inRobot.robot()->numberDof()-6, rootIsRight,leftStepArea,righttStepArea,inLeftFootShape,inRightFootShape )
{

    
    jac.resize ( 3,attNumberDof );
    attStandingRobot = &inRobot;
    robotConfiguration.resize(attNumberDof+6);

    if ( attRootIsRight )
    {
        attFrontJoint = inRobot.robot()->leftFoot();
        attRootJoint = inRobot.robot()->rightFoot();
    }
    else
    {
        attRootJoint = inRobot.robot()->leftFoot();
        attFrontJoint = inRobot.robot()->rightFoot();
    }

    matrix4d initRot,initIdealRot,invM;

    initRot = attRootJoint->currentTransformation();
    M4_IJ ( initRot,0,3 ) = 0.0;
    M4_IJ ( initRot,1,3 ) = 0.0;
    M4_IJ ( initRot,2,3 ) = 0.0;

    matrixNxP um ( 4,4 );
    vectorN euler ( 3 );
    um.clear();
    ChppGikTools::Matrix4toUblas ( initRot,um );
    ChppGikTools::RottoEulerZYX ( um,euler );
    double angle = euler ( 2 );

    MAL_S4x4_MATRIX_SET_IDENTITY ( initIdealRot );

    M4_IJ ( initIdealRot,0,0 ) = cos ( angle );
    M4_IJ ( initIdealRot,1,1 ) = M4_IJ ( initIdealRot,0,0 );
    M4_IJ ( initIdealRot,1,0 ) = sin ( angle );
    M4_IJ ( initIdealRot,0,1 ) = -M4_IJ ( initIdealRot,1,0 );

    MAL_S4x4_INVERSE ( initIdealRot, invM, double );
    attPrint2Root = invM * initRot;

    attConfiguration = subrange ( inRobot.robot()->currentConfiguration(),6,inRobot.robot()->numberDof() );

    //Preparing for velocity bound computation
    attDs.resize ( attNumberDof );
    attDi.resize ( attNumberDof );
    attUb.resize ( attNumberDof );
    attLb.resize ( attNumberDof );
    attUL.resize ( attNumberDof );
    attLL.resize ( attNumberDof );
    attMV.resize ( attNumberDof );
    att_inv_Di_Ds.resize ( attNumberDof );
    double maxVeloRatio = 0.025;
    double DiRatio = 0.10;
    double DiDsRatio = 0.5;
    for ( unsigned int i =0; i < attNumberDof; i++ )
    {
        attUL ( i ) = attStandingRobot->robot()->upperBoundDof ( i+6 );
        attLL ( i ) = attStandingRobot->robot()->lowerBoundDof ( i+6 );
        attMV ( i ) = maxVeloRatio* ( attUL ( i )-attLL ( i ) );
        attDi ( i ) = DiRatio* ( attUL ( i )-attLL ( i ) );
        attDs ( i ) = DiDsRatio*attDi ( i );
        att_inv_Di_Ds ( i ) = 1/ ( attDi ( i )-attDs ( i ) );
    }

}

void ChppLosRobotLink::setFootFreeFlyerWrtFootprint ( const matrix4d& inM )
{
    attPrint2Root = inM;
}


bool ChppLosRobotLink::upperDofLimit (unsigned int i, double ul )
{
    if (i >= attNumberDof)
    {
        std::cout << " ChppLosRobotLink::upperDofLimit wrong index" << std::endl;
        return false;
    }
    double maxVeloRatio = 0.025;
    double DiRatio = 0.10;
    double DiDsRatio = 0.5;
    attUL ( i ) = ul;
    attMV ( i ) = maxVeloRatio* ( attUL ( i )-attLL ( i ) );
    attDi ( i ) = DiRatio* ( attUL ( i )-attLL ( i ) );
    attDs ( i ) = DiDsRatio*attDi ( i );
    att_inv_Di_Ds ( i ) = 1/ ( attDi ( i )-attDs ( i ) );
    return true;
}

bool ChppLosRobotLink::lowerDofLimit (unsigned int i, double ll )
{
    if (i >= attNumberDof)
    {
        std::cout << " ChppLosRobotLink::lowerDofLimit wrong index" << std::endl;
        return false;
    }
    double maxVeloRatio = 0.025;
    double DiRatio = 0.10;
    double DiDsRatio = 0.5;
    attLL ( i ) = ll;
    attMV ( i ) = maxVeloRatio* ( attUL ( i )-attLL ( i ) );
    attDi ( i ) = DiRatio* ( attUL ( i )-attLL ( i ) );
    attDs ( i ) = DiDsRatio*attDi ( i );
    att_inv_Di_Ds ( i ) = 1/ ( attDi ( i )-attDs ( i ) );
    return true;
}

void ChppLosRobotLink::updateVelocityBounds()
{
    double dp,dn;
    for ( unsigned int i =0; i<attNumberDof;i++ )
    {

        dp = attUL ( i ) - attConfiguration ( i );
        if ( dp < attDi ( i ) )
            attUb ( i ) = attMV ( i ) * ( dp-attDs ( i ) ) *att_inv_Di_Ds ( i );
        else
            attUb ( i ) = attMV ( i );
        dn = attConfiguration ( i )-attLL ( i );
        if ( dn < attDi ( i ) )
            attLb ( i ) = -attMV ( i ) * ( dn-attDs ( i ) ) *att_inv_Di_Ds ( i );
        else
            attLb ( i ) = -attMV ( i );
    }
}

bool ChppLosRobotLink::applyConfiguration ( const ChppLosFootprint& prevleftprint, const ChppLosFootprint& prevrightprint, const vectorN& inConfiguration, bool computeRelativeFeetjacobians )
{

    if ( inConfiguration.size() < attRank + attNumberDof  )
    {
        std::cout << "ChppLosRobotLink::applyConfiguration -Error: wrong vector size" << std::endl;
        return false;
    }

    CjrlHumanoidDynamicRobot* jrlRobot = attStandingRobot->robot();

    if ( attRootIsRight )
        attRightprint =  prevrightprint;
    else
        attLeftprint = prevleftprint;

    M4_IJ ( m,0,0 ) = cos ( attRootprint->th );
    M4_IJ ( m,1,1 ) = M4_IJ ( m,0,0 );
    M4_IJ ( m,1,0 ) = sin ( attRootprint->th );
    M4_IJ ( m,0,1 ) = -M4_IJ ( m,1,0 );
    M4_IJ ( m,0,3 ) = attRootprint->x;
    M4_IJ ( m,1,3 ) = attRootprint->y;
    M4_IJ ( m,2,3 ) = jrlRobot->footHeight();

    Hif = m * attPrint2Root;



    attConfiguration = subrange(inConfiguration,attRank,attRank+attNumberDof);
    noalias(subrange(robotConfiguration,6,robotConfiguration.size())) = attConfiguration;

    {
        unsigned int i;
        for ( i =0; i< 6; i++ )
            robotConfiguration ( i ) = 0.0;

        std::vector<CjrlJoint*> supportJoints = attRootJoint->jointsFromRootToThis();
        //update joints transformations from robot root to jacobian root
        for ( i=0; i< supportJoints.size(); i++ )
            supportJoints[i]->updateTransformation ( robotConfiguration );

        //Compute robot root transformation change due to chain of joints between robot root and jacobian root
        Hf = attRootJoint->currentTransformation();
        MAL_S4x4_INVERSE(Hf,InvHf,double);
        H0 = Hif * InvHf;
        
        ChppGikTools::splitM4(H0, tmpr3d, tmpv3d);
        robotConfiguration(0) = tmpv3d(0);
        robotConfiguration(1) = tmpv3d(1);
        robotConfiguration(2) = tmpv3d(2);
        ChppGikTools::M3toEulerZYX( tmpr3d, tmpv3d);
        robotConfiguration(3) = tmpv3d(0);
        robotConfiguration(4) = tmpv3d(1);
        robotConfiguration(5) = tmpv3d(2);
        jrlRobot->currentConfiguration ( robotConfiguration );
        jrlRobot->computeForwardKinematics();
    }
    
    /*
    unsigned int frank;
    jrlRobot->currentConfiguration ( robotConfiguration );
    jrlRobot->computeForwardKinematics();
    robotConfiguration = attStandingRobot->computeConfigurationWrtFreeFlyer ( attRootJoint, mroot );

    jrlRobot->currentConfiguration ( robotConfiguration );
    jrlRobot->computeForwardKinematics();
    attConfiguration = subrange ( robotConfiguration,6,jrlRobot->numberDof() );
    */


    const matrix4d& mff = attFrontJoint->currentTransformation();
    attFrontprint->x = M4_IJ ( mff,0,3 );
    attFrontprint->y = M4_IJ ( mff,1,3 );
    attFrontprint->th = atan2 ( M4_IJ ( mff,1,0 ),M4_IJ ( mff,0,0 ) );

    attFrontprint->UXx = cos ( attFrontprint->th );
    attFrontprint->UXy = sin ( attFrontprint->th );
    attFrontprint->UYx = -attFrontprint->UXy;
    attFrontprint->UYy = attFrontprint->UXx;

    //velocity bounds
    updateVelocityBounds();

    
    if (computeRelativeFeetjacobians)
        computeJacobianPack ();
    return true;

}

void ChppLosRobotLink::computeJacobianPack ()
{
    //relative coordinates of left foot in right foot
    computeRelativeCoordinates();
    
    CjrlHumanoidDynamicRobot& robot = * ( attStandingRobot->robot() );
    CjrlJoint& rfjoint = * ( robot.rightFoot() );
    CjrlJoint& lfjoint = * ( robot.leftFoot() );
    {
        double rootth = attRightprint.th;

        double cs = cos ( rootth );
        double ss = sin ( rootth );

        robot.getPositionJacobian ( rfjoint , lfjoint, vector3d ( 0,0,0 ), jac, 0, false );
        row ( attJP.jacLinR,0 ) = cs *row ( jac,0 ) + ss *row ( jac,1 );
        row ( attJP.jacLinR,1 ) = -ss *row ( jac,0 ) + cs *row ( jac,1 );

        vector3d upperRCorner;
        upperRCorner[0] = attLeftShape.xUpRight;
        upperRCorner[1] = attLeftShape.yUpRight;
        upperRCorner[2] = -robot.footHeight();
        robot.getPositionJacobian ( rfjoint,lfjoint,upperRCorner,jac,0,false );
        row ( attJP.jacLUinR,0 ) = cs *row ( jac,0 ) + ss *row ( jac,1 );
        row ( attJP.jacLUinR,1 ) = -ss *row ( jac,0 ) + cs *row ( jac,1 );

        vector3d lowerRCorner;
        lowerRCorner[0] = attLeftShape.xDownRight;
        lowerRCorner[1] = attLeftShape.yDownRight;
        lowerRCorner[2] = -robot.footHeight();
        robot.getPositionJacobian ( rfjoint,lfjoint,lowerRCorner,jac,0,false );
        row ( attJP.jacLDinR,0 ) = cs *row ( jac,0 ) + ss *row ( jac,1 );
        row ( attJP.jacLDinR,1 ) = -ss *row ( jac,0 ) + cs *row ( jac,1 );

        robot.getOrientationJacobian ( rfjoint,lfjoint,jac,0,false );
        attJP.jacLYawinR = row ( jac,2 );
    }
    {
        double rootth = attLeftprint.th;
        double cs = cos ( rootth );
        double ss = sin ( rootth );


        robot.getPositionJacobian ( lfjoint, rfjoint, vector3d ( 0,0,0 ), jac, 0, false );
        row ( attJP.jacRinL,0 ) = cs *row ( jac,0 ) + ss *row ( jac,1 );
        row ( attJP.jacRinL,1 ) = -ss *row ( jac,0 ) + cs *row ( jac,1 );

        vector3d upperLCorner;
        upperLCorner[0] = attRightShape.xUpLeft;
        upperLCorner[1] = attRightShape.yUpLeft;
        upperLCorner[2] = -robot.footHeight();
        robot.getPositionJacobian ( lfjoint,rfjoint,upperLCorner,jac,0,false );
        row ( attJP.jacRUinL,0 ) = cs *row ( jac,0 ) + ss *row ( jac,1 );
        row ( attJP.jacRUinL,1 ) = -ss *row ( jac,0 ) + cs *row ( jac,1 );

        vector3d lowerLCorner;
        lowerLCorner[0] = attRightShape.xDownLeft;
        lowerLCorner[1] = attRightShape.yDownLeft;
        lowerLCorner[2] = -robot.footHeight();
        robot.getPositionJacobian ( lfjoint,rfjoint,lowerLCorner,jac,0,false );
        row ( attJP.jacRDinL,0 ) = cs *row ( jac,0 ) + ss *row ( jac,1 );
        row ( attJP.jacRDinL,1 ) = -ss *row ( jac,0 ) + cs *row ( jac,1 );

        robot.getOrientationJacobian ( lfjoint,rfjoint,jac,0,false );
        attJP.jacRYawinL = row ( jac,2 );
    }
}


bool ChppLosRobotLink::getJacobian ( const vector3d& inPoint, matrixNxP& outJacobian, unsigned int inRank )
{
    if ( ( outJacobian.size2() < inRank +attNumberDof ) || ( outJacobian.size1() !=6 ) )
    {
        std::cout << "ChppLosRobotLink::getJacobian -Error bad size for outJacobian " << std::endl;
        return false;
    }
    vector3d vecL ( 0,0,inPoint[2] );
    double dx = inPoint[0] - attFrontprint->x;
    double dy = inPoint[1] - attFrontprint->y;

    vecL[0] = cos ( attFrontprint->th ) *dx + sin ( attFrontprint->th ) *dy;
    vecL[1] = -sin ( attFrontprint->th ) *dx + cos ( attFrontprint->th ) *dy;

    attStandingRobot->robot()->getJacobian ( *attRootJoint, *attFrontJoint, vecL, outJacobian, inRank, false );

    return true;
}

bool ChppLosRobotLink::getPositionJacobian ( const vector3d& inPoint, matrixNxP& outJacobian, unsigned int inRank )
{
    if ( ( outJacobian.size2() < inRank +attNumberDof ) || ( outJacobian.size1() !=3 ) )
    {
        std::cout << "ChppLosRobotLink::getJacobian -Error bad size for outJacobian " << std::endl;
        return false;
    }
    vector3d vecL ( 0,0,inPoint[2] );
    double dx = inPoint[0] - attFrontprint->x;
    double dy = inPoint[1] - attFrontprint->y;

    vecL[0] = cos ( attFrontprint->th ) *dx + sin ( attFrontprint->th ) *dy;
    vecL[1] = -sin ( attFrontprint->th ) *dx + cos ( attFrontprint->th ) *dy;

    attStandingRobot->robot()->getPositionJacobian ( *attRootJoint, *attFrontJoint, vecL, outJacobian, inRank, false );

    return true;
}

bool ChppLosRobotLink::getOrientationJacobian ( matrixNxP& outJacobian, unsigned int inRank )
{
    if ( ( outJacobian.size2() < inRank +attNumberDof ) || ( outJacobian.size1() !=3 ) )
    {
        std::cout << "ChppLosRobotLink::getJacobian -Error bad size for outJacobian " << std::endl;
        return false;
    }

    attStandingRobot->robot()->getOrientationJacobian ( *attRootJoint, *attFrontJoint, outJacobian, inRank, false );

    return true;
}

void ChppLosRobotLink::rootIsRight ( bool newRootIsRight )
{
    if ( newRootIsRight != attRootIsRight )
    {
        attRootIsRight = newRootIsRight;
        
        if ( attRootIsRight )
        {
            attFrontJoint = attStandingRobot->robot()->leftFoot();
            attRootJoint = attStandingRobot->robot()->rightFoot();
            attRootprint = &attRightprint;
            attFrontprint = &attLeftprint;
        }
        else
        {
            attRootJoint = attStandingRobot->robot()->leftFoot();
            attFrontJoint = attStandingRobot->robot()->rightFoot();
            attRootprint = &attLeftprint;
            attFrontprint = &attRightprint;
        }

        matrix4d initRot,initIdealRot,invM;

        initRot = attRootJoint->currentTransformation();
        M4_IJ ( initRot,0,3 ) = 0.0;
        M4_IJ ( initRot,1,3 ) = 0.0;
        M4_IJ ( initRot,2,3 ) = 0.0;

        matrixNxP um ( 4,4 );
        vectorN euler ( 3 );
        um.clear();
        ChppGikTools::Matrix4toUblas ( initRot,um );
        ChppGikTools::RottoEulerZYX ( um,euler );
        double angle = euler ( 2 );

        MAL_S4x4_MATRIX_SET_IDENTITY ( initIdealRot );

        M4_IJ ( initIdealRot,0,0 ) = cos ( angle );
        M4_IJ ( initIdealRot,1,1 ) = M4_IJ ( initIdealRot,0,0 );
        M4_IJ ( initIdealRot,1,0 ) = sin ( angle );
        M4_IJ ( initIdealRot,0,1 ) = -M4_IJ ( initIdealRot,1,0 );

        MAL_S4x4_INVERSE ( initIdealRot, invM, double );
        attPrint2Root = invM * initRot;
    }
}

CjrlJoint* ChppLosRobotLink::rootJoint()
{
    return attRootJoint;
}
CjrlJoint* ChppLosRobotLink::frontJoint()
{
    return attFrontJoint;
}
