/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpplocalstepper/hppLosSolverBasic.hh"
#include "hpphik/hppHikSolverBasic.hh"
#include "hpplocalstepper/tasks/hppLosTask.hh"
#include "hpplocalstepper/hppLosCentipede.hh"
#include "hppGik/hppGikTools.h"

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J
#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J

using namespace boost::numeric::ublas;

ChppLosSolverBasic::ChppLosSolverBasic ( ChppLosCentipede& inC )
{
    attCentipede = &inC;

    attHikSolver = new ChppHikSolverBasic ( attCentipede->numberDof() );

    attActivatedJoints = scalar_vector<double> ( attCentipede->numberDof(), 1 );
    attHikSolver->setMask ( attActivatedJoints );

    attSolution.resize ( attCentipede->numberDof() );
    attUb.resize ( attCentipede->numberDof() );
    attLb.resize ( attCentipede->numberDof() );
}

       
ChppLosSolverBasic::~ChppLosSolverBasic()
{
    delete attHikSolver;
}

void ChppLosSolverBasic::tasks ( const std::vector<ChppLosTask*>& inTasks )
{
    attTasks = inTasks;
}

void ChppLosSolverBasic::constraints ( const std::vector<ChppLosTask*>& inConstraints )
{
    attConstraints = inConstraints;
}

bool ChppLosSolverBasic::setMask ( const vectorN& inVector )
{
    if ( inVector.size() != attCentipede->numberDof() )
    {
        std::cout << "ChppLosSolverBasic::setMask -Error bad vector size"<< std::endl;
        return false;
    }

    attActivatedJoints = inVector;
    
    if (attHikSolver->numberParams() != attCentipede->numberDof())
       resize();

    attHikSolver->setMask ( attActivatedJoints );
    return true;
}

void ChppLosSolverBasic::resize()
{
    delete attHikSolver;
    attHikSolver = new ChppHikSolverBasic ( attCentipede->numberDof() );
    attUb.resize( attCentipede->numberDof() );
    attLb.resize( attCentipede->numberDof() );
}

bool ChppLosSolverBasic::solve(vectorN& veloSolution)
{
    int retV;
    unsigned int i;
    
    if (veloSolution.size() != attCentipede->numberDof())
    {
        std::cout << "ChppLosSolverBasic::solve Bad size of initial guess" << std::endl;
        return false;
    }
        
    if (attHikSolver->numberParams() != attCentipede->numberDof())
        resize();
    
    //Set parameter bounds
    attHikSolver->clearBounds();
    attCentipede->getUpperVelocityBounds(attUb);
    attCentipede->getLowerVelocityBounds(attLb);
    for ( i=0;i<attCentipede->numberDof();i++ )
    {
        attHikSolver->upperBound ( i, attUb( i ) );
        attHikSolver->lowerBound ( i, attLb( i ) );
    }

    //Set initial feasible set
    loadConstraintsInSolver();


    for ( i = 0; i< attTasks.size(); i++ )
    {
        attHikSolver->clearTasks();
        attTasks[i]->computeConstraint();
        if ( attTasks[i]->isEqualityTask() )
            attHikSolver->addTask ( attTasks[i]->constraint(), true );
        else
            attHikSolver->addTask ( attTasks[i]->constraint(), false );

        retV = attHikSolver->solve ( veloSolution, attTasks[i]->dampingFactor() );
        if ( retV !=0 )
        {
            std::cout << "retV = " << retV << " at task ranked "<< i<< std::endl;
            return false;
        }

        if ( i!= attTasks.size()-1 )
            attHikSolver->updateFeasibleSet();
    }
    veloSolution = attHikSolver->solution();
    
    return true;
}


void ChppLosSolverBasic::loadConstraintsInSolver()
{
    attHikSolver->removePreConstraints();

    for ( unsigned int i=0;i<attConstraints.size();i++ )
    {
        attConstraints[i]->computeConstraint();
        if ( attConstraints[i]->isEqualityTask() )
            attHikSolver->addEqPreConstraint ( attConstraints[i]->constraint() );
        else
            attHikSolver->addIneqPreConstraint ( attConstraints[i]->constraint() );
    }
    attHikSolver->initFeasibleSet();
}

