/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "hpplocalstepper/tasks/hppLosCollisionAvoider.hh"
#include <string.h>

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

void ChppLosCollisionAvoider::crossProd3 ( const vector3d& v1, const vector3d& v2, vector3d& retV )
{
    retV ( 0 ) = v1 ( 1 ) *v2 ( 2 ) - v1 ( 2 ) *v2 ( 1 );
    retV ( 1 ) = v1 ( 2 ) *v2 ( 0 ) - v1 ( 0 ) *v2 ( 2 );
    retV ( 2 ) = v1 ( 0 ) *v2 ( 1 ) - v1 ( 1 ) *v2 ( 0 );
}

ChppLosCollisionAvoider::ChppLosCollisionAvoider ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inLink ) :ChppLosTask ( inCentipede )
{
    attSolver = new ChppHikSlackSolver ( 2 );
    attSolver->lowerBound ( 0,0 );
    attSolver->lowerBound ( 1,0 );
    attSolver->upperBound ( 0,1 );
    attSolver->upperBound ( 1,1 );
    attSystem = new ChppHikLinearSystem ( 2 );
    attSystem->matrix().resize ( 3,2 );
    attSystem->vector().resize ( 3 );

    attConstraint->vector().resize ( 0 );
    attConstraint->matrix().resize ( 0, attCentipede->numberDof() );
    attValue.resize ( 0 );

    noutM.resize ( 1,3 );

    tempJac3d.resize ( 3, inLink.numberDof() );
    tempJac3d.clear();
    tempJac3dLong.resize ( 3, attCentipede->numberDof() );
    tempJac3dLong.clear();
    jac1d.resize ( 1, inLink.numberDof() );
    jac1d.clear();

    attDi = 0.05;
    attDs = 0.02;
    attXi = 0.02;
}

ChppLosCollisionAvoider::~ChppLosCollisionAvoider()
{
    delete attSolver;
    delete attSystem;
}

void ChppLosCollisionAvoider::setMargins ( double inDi, double inDs, double inXi )
{
    attXi = inXi;
    attDs = inDs;
    attDi = inDi;
}


unsigned int ChppLosCollisionAvoider::addRobotSphere ( LosRobotSphere* inS )
{
    attRobotSpheres.push_back ( inS );
    return attRobotSpheres.size()-1;
}

unsigned int ChppLosCollisionAvoider::addExtPlane ( LosExtPlane* inP )
{
    attExtPlanes.push_back ( inP );
    return attExtPlanes.size()-1;
}

unsigned int ChppLosCollisionAvoider::addRobotCylinder ( LosRobotCylinder* inC )
{
    attRobotCylinders.push_back ( inC );
    return attRobotCylinders.size()-1;
}

unsigned int ChppLosCollisionAvoider::addLinkFootprint ( LosLinkFootprint* inC )
{
    attLinkFootprints.push_back ( inC );
    return attLinkFootprints.size()-1;
}

unsigned int ChppLosCollisionAvoider::addExtCylinder ( LosExtCylinder* inC )
{
    attExtCylinders.push_back ( inC );
    return attExtCylinders.size()-1;
}

bool ChppLosCollisionAvoider::makeRobotCylinderPair ( unsigned int id_1, unsigned int id_2 )
{
    if ( id_1 == id_2 )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderPair Indices equal"<<std::endl;
        return false;
    }
    if ( ( id_1 >= attRobotCylinders.size() ) || ( id_2 >= attRobotCylinders.size() ) )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderPair Bad indices"<<std::endl;
        return false;
    }
    if ( attRobotCylinders[id_2]->link != attRobotCylinders[id_1]->link )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderPair not the same robot link"<<std::endl;
        return false;
    }
    attRobotCylinderPairs.push_back ( std::make_pair ( attRobotCylinders[id_1],attRobotCylinders[id_2] ) );
    SelfDamper Dp;
    Dp.link = attRobotCylinders[id_1]->link;
    Dp.joint2 = attRobotCylinders[id_2]->joint;
    Dp.joint1 = attRobotCylinders[id_1]->joint;
    Dp.Ds = attRobotCylinders[id_1]->radius + attRobotCylinders[id_2]->radius + attDs;
    Dp.Di = attRobotCylinders[id_1]->radius + attRobotCylinders[id_2]->radius + attDi;
    Dp.Di_Ds = attDi - attDs;
    attRobotCylinderDampers.push_back ( Dp );
    return true;
}


bool ChppLosCollisionAvoider::makeRobotCylinderPlanePair ( unsigned int id_1, unsigned int id_2 )
{
    if ( ( id_1 >= attRobotCylinders.size() ) || ( id_2 >= attExtPlanes.size() ) )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderExtPlanePair Bad indices"<<std::endl;
        return false;
    }
    attRobotCylinderPlanePairs.push_back ( std::make_pair ( attRobotCylinders[id_1],attExtPlanes[id_2] ) );
    ExtDamper Dp;
    Dp.link = attRobotCylinders[id_1]->link;
    Dp.joint = attRobotCylinders[id_1]->joint;
    Dp.ntojoint = attExtPlanes[id_2]->N;
    Dp.Ds = attRobotCylinders[id_1]->radius + attExtPlanes[id_2]->ds + attDs;
    Dp.Di = attRobotCylinders[id_1]->radius + attExtPlanes[id_2]->ds + attDi;
    Dp.Di_Ds = attDi - attDs;

    attRobotCylinderPlaneDampers.push_back ( Dp );
    return true;
}

bool ChppLosCollisionAvoider::makeLinkFootprintPlanePair ( unsigned int id_1, unsigned int id_2 )
{
    if ( ( id_1 >= attLinkFootprints.size() ) || ( id_2 >= attExtPlanes.size() ) )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderExtPlanePair Bad indices"<<std::endl;
        return false;
    }
    attLinkFootprintPlanePairs.push_back ( std::make_pair ( attLinkFootprints[id_1],attExtPlanes[id_2] ) );
    LinkDamper Dp;
    Dp.link = attLinkFootprints[id_1]->link;
    Dp.ntojoint = attExtPlanes[id_2]->N;
    Dp.Ds = attLinkFootprints[id_1]->ds + attExtPlanes[id_2]->ds + attDs;
    Dp.Di = attLinkFootprints[id_1]->ds + attExtPlanes[id_2]->ds + attDi;
    Dp.Di_Ds = attDi - attDs;

    attLinkFootprintPlaneDampers.push_back ( Dp );
    return true;
}

bool ChppLosCollisionAvoider::makeRobotCylinderExtPair ( unsigned int id_1, unsigned int id_2 )
{
    if ( ( id_1 >= attRobotCylinders.size() ) || ( id_2 >= attExtCylinders.size() ) )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderExtPair Bad indices"<<std::endl;
        return false;
    }
    attRobotCylinderExtPairs.push_back ( std::make_pair ( attRobotCylinders[id_1],attExtCylinders[id_2] ) );
    ExtDamper Dp;
    Dp.link = attRobotCylinders[id_1]->link;
    Dp.joint = attRobotCylinders[id_1]->joint;
    Dp.Ds = attRobotCylinders[id_1]->radius + attExtCylinders[id_2]->radius + attDs;
    Dp.Di = attRobotCylinders[id_1]->radius + attExtCylinders[id_2]->radius + attDi;
    Dp.Di_Ds = attDi - attDs;
    attRobotCylinderExtDampers.push_back ( Dp );
    return true;
}

bool ChppLosCollisionAvoider::makeRobotCylinderSpherePair ( unsigned int id_1, unsigned int id_2 )
{
    if ( ( id_1 >= attRobotCylinders.size() ) || ( id_2 >= attRobotSpheres.size() ) )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderSpherePair Bad indices"<<std::endl;
        return false;
    }
    if ( attRobotSpheres[id_2]->link != attRobotCylinders[id_1]->link )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotCylinderSpherePair not the same robot link"<<std::endl;
        return false;
    }
    attRobotCylinderSpherePairs.push_back ( std::make_pair ( attRobotCylinders[id_1],attRobotSpheres[id_2] ) );
    SelfDamper Dp;
    Dp.link = attRobotCylinders[id_1]->link;
    Dp.joint2 = attRobotSpheres[id_2]->joint;
    Dp.joint1 = attRobotCylinders[id_1]->joint;
    Dp.Ds = attRobotCylinders[id_1]->radius + attRobotSpheres[id_2]->radius + attDs;
    Dp.Di = attRobotCylinders[id_1]->radius + attRobotSpheres[id_2]->radius + attDi;
    Dp.Di_Ds = attDi - attDs;
    Dp.locPoint2 = attRobotSpheres[id_2]->locS;
    attRobotCylinderSphereDampers.push_back ( Dp );
    return true;
}

bool ChppLosCollisionAvoider::makeRobotSpherePair ( unsigned int id_1, unsigned int id_2 )
{
    if ( id_1 == id_2 )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotSpherePair Indices equal"<<std::endl;
        return false;
    }
    if ( ( id_1 >= attRobotSpheres.size() ) || ( id_2 >= attRobotSpheres.size() ) )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotSpherePair Bad indices"<<std::endl;
        return false;
    }
    if ( attRobotSpheres[id_2]->link != attRobotSpheres[id_1]->link )
    {
        std::cout<< "ChppLosCollisionAvoider::makeRobotSpherePair not the same robot link"<<std::endl;
        return false;
    }

    attRobotSpherePairs.push_back ( std::make_pair ( attRobotSpheres[id_1],attRobotSpheres[id_2] ) );
    SelfDamper Dp;
    Dp.link = attRobotSpheres[id_1]->link;
    Dp.joint2 = attRobotSpheres[id_2]->joint;
    Dp.joint1 = attRobotSpheres[id_1]->joint;
    Dp.Ds = attRobotSpheres[id_1]->radius + attRobotSpheres[id_2]->radius + attDs;
    Dp.Di = attRobotSpheres[id_1]->radius + attRobotSpheres[id_2]->radius + attDi;
    Dp.Di_Ds = attDi - attDs;
    Dp.locPoint2 = attRobotSpheres[id_2]->locS;
    Dp.locPoint1 = attRobotSpheres[id_1]->locS;
    attRobotSphereDampers.push_back ( Dp );
    return true;
}


bool ChppLosCollisionAvoider::checkRobotSpherePair ( unsigned int rank )
{
    std::pair< LosRobotSphere*, LosRobotSphere* > &pr = attRobotSpherePairs[rank];
    SelfDamper& dp = attRobotSphereDampers[rank];

    vector3d diff = pr.second->S - pr.first->S;
    dp.distance = sqrt ( diff[0]*diff[0] + diff[1]*diff[1] + diff[2]*diff[2] );

    if ( dp.distance <  dp.Di );
    {
        dp.n12 = diff/dp.distance;
        return true;
    }
    return false;
}

bool ChppLosCollisionAvoider::checkRobotCylinderSpherePair ( unsigned int rank )
{
    std::pair< LosRobotCylinder*, LosRobotSphere* > &pr = attRobotCylinderSpherePairs[rank];
    SelfDamper& dp = attRobotCylinderSphereDampers[rank];

    vector3d diff = pr.second->S - pr.first->A;
    vector3d uAB =  pr.first->AB/pr.first->normAB;
    double coef = ( diff[0]*uAB[0] + diff[1]*uAB[1] + diff[2]*uAB[2] ) /pr.first->normAB;
    if ( coef <0 )
    {
        dp.distance = sqrt ( diff[0]*diff[0] + diff[1]*diff[1] + diff[2]*diff[2] );
        if ( dp.distance < dp.Di )
        {
            dp.n12 = diff/dp.distance;
            dp.locPoint1 = pr.first->locA;
            return true;
        }
        return false;
    }
    if ( coef >1 )
    {
        diff += pr.first->AB;
        dp.distance = sqrt ( diff[0]*diff[0] + diff[1]*diff[1] + diff[2]*diff[2] );
        if ( dp.distance < dp.Di )
        {
            dp.n12 = diff/dp.distance;
            dp.locPoint1 = pr.first->locB;
            return true;
        }
        return false;
    }

    diff += pr.first->AB* ( -coef );
    dp.distance = sqrt ( diff[0]*diff[0] + diff[1]*diff[1] + diff[2]*diff[2] );
    if ( dp.distance < dp.Di )
    {
        dp.n12 = diff/dp.distance;
        dp.locPoint1 = pr.first->locA* ( 1-coef ) + pr.first->locB*coef;
        return true;
    }
    return false;
}


bool ChppLosCollisionAvoider::checkRobotCylinderPlanePair ( unsigned int rank )
{
    std::pair< LosRobotCylinder*, LosExtPlane* > &pr = attRobotCylinderPlanePairs[rank];
    ExtDamper& dp = attRobotCylinderPlaneDampers[rank];


    vector3d diff =  pr.first->A-pr.second->P;
    double dist1 =  pr.second->N[0]*diff[0] + pr.second->N[1]*diff[1] + pr.second->N[2]*diff[2];

    diff =  pr.first->B-pr.second->P;
    double dist2 =  pr.second->N[0]*diff[0] + pr.second->N[1]*diff[1] + pr.second->N[2]*diff[2];


    if ( dist1 < dist2 )
    {
        dp.distance = dist1;
        dp.locPoint = pr.first->locA;
        dp.absPoint = pr.first->A;
    }
    else
    {
        dp.distance = dist2;
        dp.locPoint = pr.first->locB;
        dp.absPoint = pr.first->B;
    }

    if ( dp.distance < dp.Di )
    {
        //std::cout << " dp.distance " << dp.distance<<std::endl;
        return true;
    }

    return false;
}


bool ChppLosCollisionAvoider::checkLinkFootprintPlanePair ( unsigned int rank )
{
    std::pair< LosLinkFootprint*, LosExtPlane* > &pr = attLinkFootprintPlanePairs[rank];
    LinkDamper& dp = attLinkFootprintPlaneDampers[rank];


    vector3d* vects[4];
    double dists[4];

    vector3d diff =  pr.first->P1-pr.second->P;
    double dist1 =  pr.second->N[0]*diff[0] + pr.second->N[1]*diff[1] + pr.second->N[2]*diff[2];
    vects[0] = &pr.first->P1;
    dists[0] = dist1;

    diff =  pr.first->P2-pr.second->P;
    double dist2 =  pr.second->N[0]*diff[0] + pr.second->N[1]*diff[1] + pr.second->N[2]*diff[2];
    vects[1] = &pr.first->P2;
    dists[1] = dist2;

    diff =  pr.first->P3-pr.second->P;
    double dist3 =  pr.second->N[0]*diff[0] + pr.second->N[1]*diff[1] + pr.second->N[2]*diff[2];
    vects[2] = &pr.first->P3;
    dists[2] = dist3;

    diff =  pr.first->P4-pr.second->P;
    double dist4 =  pr.second->N[0]*diff[0] + pr.second->N[1]*diff[1] + pr.second->N[2]*diff[2];
    vects[3] = &pr.first->P4;
    dists[3] = dist4;

    int j = 0;
    while ( j<3 )
    {
        bool no_change = true;
        vector3d* swapVec;
        double swapVal;
        for ( unsigned int i=0;i<3;i++ )
            if ( dists[i] > dists[i+1] )
            {
                swapVec = vects[i+1];
                swapVal = dists[i+1];
                no_change = false;
                dists[i+1] = dists[i];
                vects[i+1] = vects[i];
                dists[i] = swapVal;
                vects[i] = swapVec;
            }
        if ( no_change )
            break;
        j++;
    }


    dp.distance1 = dists[0];
    dp.absPoint1 = *vects[0];

    dp.distance2 = dists[1];
    dp.absPoint2 = *vects[1];


    if ( dp.distance1 < dp.Di )
        return true;

    return false;
}


bool ChppLosCollisionAvoider::checkRobotCylinderExtPair ( unsigned int rank )
{
    std::pair< LosRobotCylinder*, LosExtCylinder* > &pr = attRobotCylinderExtPairs[rank];
    ExtDamper& dp = attRobotCylinderExtDampers[rank];


    attSystem->matrix() ( 0,0 ) = -pr.second->AB[0];
    attSystem->matrix() ( 1,0 ) = -pr.second->AB[1];
    attSystem->matrix() ( 2,0 ) = -pr.second->AB[2];

    attSystem->matrix() ( 0,1 ) = pr.first->AB[0];
    attSystem->matrix() ( 1,1 ) = pr.first->AB[1];
    attSystem->matrix() ( 2,1 ) = pr.first->AB[2];

    attSystem->vector() ( 0 ) = pr.second->A[0] - pr.first->A[0];
    attSystem->vector() ( 1 ) = pr.second->A[1] - pr.first->A[1];
    attSystem->vector() ( 2 ) = pr.second->A[2] - pr.first->A[2];

    attSolver->clearTasks();
    attSolver->addTasks ( attSystem,true );
    vectorN lambda ( 2 ),sv;
    attSolver->solve ( lambda,sv );

    vector3d diff = pr.first->A - pr.second->A +  pr.first->AB*lambda ( 1 ) - pr.second->AB*lambda ( 0 );

    dp.distance = sqrt ( diff[0]*diff[0] + diff[1]*diff[1] + diff[2]*diff[2] );
    if ( dp.distance < dp.Di )
    {
        dp.ntojoint = diff/dp.distance;
        dp.locPoint =  pr.first->locA* ( 1-lambda ( 1 ) ) + pr.first->locB*lambda ( 1 );
        dp.absPoint = pr.first->A* ( 1-lambda ( 1 ) ) + pr.first->B*lambda ( 1 );
        return true;
    }

    return false;
}

bool ChppLosCollisionAvoider::checkRobotCylinderPair ( unsigned int rank )
{
    std::pair< LosRobotCylinder*, LosRobotCylinder *> &pr = attRobotCylinderPairs[rank];
    SelfDamper& dp = attRobotCylinderDampers[rank];

    /*
    vector3d cp;
    crossProd3 ( pr.first.AB, pr.second.AB, cp );
    double ncp = S3_VECTOR_NORM(cp);
    */

    attSystem->matrix() ( 0,0 ) = -pr.first->AB[0];
    attSystem->matrix() ( 1,0 ) = -pr.first->AB[1];
    attSystem->matrix() ( 2,0 ) = -pr.first->AB[2];

    attSystem->matrix() ( 0,1 ) = pr.second->AB[0];
    attSystem->matrix() ( 1,1 ) = pr.second->AB[1];
    attSystem->matrix() ( 2,1 ) = pr.second->AB[2];

    attSystem->vector() ( 0 ) = pr.first->A[0] - pr.second->A[0];
    attSystem->vector() ( 1 ) = pr.first->A[1] - pr.second->A[1];
    attSystem->vector() ( 2 ) = pr.first->A[2] - pr.second->A[2];

    attSolver->clearTasks();
    attSolver->addTasks ( attSystem,true );
    vectorN lambda ( 2 ),sv;
    attSolver->solve ( lambda,sv );

    vector3d diff = pr.second->A - pr.first->A +  pr.second->AB*lambda ( 1 ) - pr.first->AB*lambda ( 0 );

    dp.distance = sqrt ( diff[0]*diff[0] + diff[1]*diff[1] + diff[2]*diff[2] );
    
    if ( dp.distance < dp.Di )
    {
        dp.n12 = diff/dp.distance;
        dp.locPoint1 =  pr.first->locA* ( 1-lambda ( 0 ) ) + pr.first->locB*lambda ( 0 );
        dp.locPoint2 = pr.second->locA* ( 1-lambda ( 1 ) ) + pr.second->locB*lambda ( 1 );
        
        
        //std::cout << "cyl-cyl ranked " << rank << " dp.distance " << dp.distance << "  Di " <<  dp.Di << std::endl;
        
        vector3d par;
        crossProd3(pr.second->AB,pr.first->AB,par);
        //std::cout << sqrt(par[0]*par[0]+ par[1]*par[1] + par[2]*par[2]) << std::endl;
        
        return true;
    }
    return false;
}


void ChppLosCollisionAvoider::calcConstraint ( SelfDamper& inDamper, unsigned int constraintRank )
{


    unsigned int startRank = inDamper.link->rank();

    tempJac3d.clear();
    inDamper.link->standingRobot().robot()->getPositionJacobian ( *inDamper.joint1, *inDamper.joint2, inDamper.locPoint2, tempJac3d, 0, false );

    noutM ( 0,0 ) = inDamper.n12[0];
    noutM ( 0,1 ) = inDamper.n12[1];
    noutM ( 0,2 ) = inDamper.n12[2];

    noalias ( jac1d ) = prod ( -noutM,tempJac3d );
    memcpy ( ( &attConstraint->matrix().data() [constraintRank*attCentipede->numberDof() +startRank] ),&jac1d.data() [0],inDamper.link->numberDof() *sizeof ( double ) );


    attConstraint->vector() ( constraintRank ) = attXi* ( inDamper.distance - inDamper.Ds ) / inDamper.Di_Ds ;

    /*
    if ( dist < attCRadius + attDi )
        attConstraint->vector() ( rank ) = attXi* ( dist - attCRadius - attDs ) / ( attDi - attDs );
    else
        attConstraint->vector() ( rank ) = attXi;
    */
}

void ChppLosCollisionAvoider::calcExtConstraint ( ExtDamper& inDamper, unsigned int constraintRank )
{
    //std::cout << "calculating! "<<std::endl;

    unsigned int startRank = inDamper.link->rank();

    tempJac3dLong.clear();
    inDamper.link->standingRobot().robot()->getPositionJacobian ( *inDamper.link->rootJoint(), *inDamper.joint, inDamper.locPoint, tempJac3dLong, startRank, false );
    //std::cout << inDamper.joint<< " " << " locP " <<  inDamper.locPoint << " absP " <<  inDamper.absPoint << std::endl;
    ChppLosLink* plink = inDamper.link->parent();
    while ( plink!=0 )
    {
        //std::cout << plink->rank() << std::endl;
        plink->getPositionJacobian ( inDamper.absPoint,tempJac3dLong,plink->rank() );
        plink = plink->parent();
    }

    //std::cout << tempJac3dLong << std::endl;
    
    noutM ( 0,0 ) = inDamper.ntojoint[0];
    noutM ( 0,1 ) = inDamper.ntojoint[1];
    noutM ( 0,2 ) = inDamper.ntojoint[2];

    noalias ( subrange ( attConstraint->matrix(),constraintRank,constraintRank+1,0,attCentipede->numberDof() ) ) = prod ( -noutM,tempJac3dLong );

    attConstraint->vector() ( constraintRank ) = attXi* ( inDamper.distance - inDamper.Ds ) / inDamper.Di_Ds ;
    //std::cout << attConstraint->vector() ( constraintRank ) << std::endl;
}


void ChppLosCollisionAvoider::calcLinkConstraint ( LinkDamper& inDamper, unsigned int constraintRank )
{
    tempJac3dLong.clear();
    ChppLosLink* plink = inDamper.link->parent();
    while ( plink!=0 )
    {
        plink->getPositionJacobian ( inDamper.absPoint1,tempJac3dLong,plink->rank() );
        plink = plink->parent();
    }

    noutM ( 0,0 ) = inDamper.ntojoint[0];
    noutM ( 0,1 ) = inDamper.ntojoint[1];
    noutM ( 0,2 ) = inDamper.ntojoint[2];

    noalias ( subrange ( attConstraint->matrix(),constraintRank,constraintRank+1,0,attCentipede->numberDof() ) ) = prod ( -noutM,tempJac3dLong );

    attConstraint->vector() ( constraintRank ) = attXi* ( inDamper.distance1 - inDamper.Ds ) / inDamper.Di_Ds ;


    plink = inDamper.link->parent();
    while ( plink!=0 )
    {
        plink->getPositionJacobian ( inDamper.absPoint2,tempJac3dLong,plink->rank() );
        plink = plink->parent();
    }

    noalias ( subrange ( attConstraint->matrix(),constraintRank+1,constraintRank+2,0,attCentipede->numberDof() ) ) = prod ( -noutM,tempJac3dLong );

    attConstraint->vector() ( constraintRank+1 ) = attXi* ( inDamper.distance2 - inDamper.Ds ) / inDamper.Di_Ds ;

}

void ChppLosCollisionAvoider::updateModel()
{
    matrix4d ma;
    for ( unsigned int i = 0;i<attRobotCylinders.size();i++ )
    {
        ma = attRobotCylinders[i]->joint->currentTransformation();
        attRobotCylinders[i]->A =ma*attRobotCylinders[i]->locA;
        attRobotCylinders[i]->B =ma*attRobotCylinders[i]->locB;
        attRobotCylinders[i]->AB =  attRobotCylinders[i]->B - attRobotCylinders[i]->A;
    }



    for ( unsigned int i = 0;i<attRobotSpheres.size();i++ )
    {
        ma = attRobotSpheres[i]->joint->currentTransformation();
        attRobotSpheres[i]->S =ma*attRobotSpheres[i]->locS;
    }

    double offsety  = 0.15,offsetx = 0.1;
    for ( unsigned int i = 0;i<attLinkFootprints.size();i++ )
    {
        const ChppLosFootprint& fp = attLinkFootprints[i]->link->frontprint();
        const ChppLosFootprintShape* shp;
        if ( attLinkFootprints[i]->link->rootIsRight() )
            shp = &attLinkFootprints[i]->link->leftShapeArea();
        else
            shp = &attLinkFootprints[i]->link->rightShapeArea();

        attLinkFootprints[i]->P1 ( 0 ) = fp.x + fp.UXx*(shp->xUpRight + offsetx) + fp.UYx*(shp->yUpRight - offsety);
        attLinkFootprints[i]->P1 ( 1 ) = fp.y + fp.UXy*(shp->xUpRight + offsetx) + fp.UYy*(shp->yUpRight - offsety);
        attLinkFootprints[i]->P1 ( 2 ) = 0.0;

        attLinkFootprints[i]->P2 ( 0 ) = fp.x + fp.UXx*(shp->xDownRight - offsetx) + fp.UYx*(shp->yDownRight - offsety);
        attLinkFootprints[i]->P2 ( 1 ) = fp.y + fp.UXy*(shp->xDownRight - offsetx) + fp.UYy*(shp->yDownRight - offsety);
        attLinkFootprints[i]->P2 ( 2 ) = 0.0;

        attLinkFootprints[i]->P3 ( 0 ) = fp.x + fp.UXx*(shp->xDownLeft - offsetx) + fp.UYx*(shp->yDownLeft+ offsety);
        attLinkFootprints[i]->P3 ( 1 ) = fp.y + fp.UXy*(shp->xDownLeft - offsetx) + fp.UYy*(shp->yDownLeft+ offsety);
        attLinkFootprints[i]->P3 ( 2 ) = 0.0;

        attLinkFootprints[i]->P4 ( 0 ) = fp.x + fp.UXx*(shp->xUpLeft + offsetx) + fp.UYx*(shp->yUpLeft+ offsety);
        attLinkFootprints[i]->P4 ( 1 ) = fp.y + fp.UXy*(shp->xUpLeft + offsetx) + fp.UYy*(shp->yUpLeft+ offsety);
        attLinkFootprints[i]->P4 ( 2 ) = 0.0;
    }


}

void ChppLosCollisionAvoider::computeConstraint()
{
    resize();

    updateModel();

    std::vector<SelfDamper*> dampers;
    std::vector<ExtDamper*> extDampers;
    std::vector<LinkDamper*> linkDampers;
    for ( unsigned int i = 0;i<attRobotCylinderSphereDampers.size();i++ )
        if ( checkRobotCylinderSpherePair ( i ) )
        {
            //std::cout << "active pair of sphere-cylinder at rank " << i <<std::endl;
            dampers.push_back ( &attRobotCylinderSphereDampers[i] );
        }

    for ( unsigned int i = 0;i<attRobotCylinderDampers.size();i++ )
        if ( checkRobotCylinderPair ( i ) )
        {
            //std::cout << "active pair of cylinders at rank " << i <<std::endl;
            dampers.push_back ( &attRobotCylinderDampers[i] );
        }

    for ( unsigned int i = 0;i<attRobotSphereDampers.size();i++ )
        if ( checkRobotSpherePair ( i ) )
        {
            //std::cout << "active pair of spheres at rank " << i <<std::endl;
            dampers.push_back ( &attRobotSphereDampers[i] );
        }

    for ( unsigned int i = 0;i<attRobotCylinderPlaneDampers.size();i++ )
        if ( checkRobotCylinderPlanePair ( i ) )
        {
            //std::cout << "active pair of Cylinder-plane at rank " << i <<std::endl;
            extDampers.push_back ( &attRobotCylinderPlaneDampers[i] );
        }

    for ( unsigned int i = 0;i<attRobotCylinderExtDampers.size();i++ )
        if ( checkRobotCylinderExtPair ( i ) )
        {
            //std::cout << "active pair of Cylinder-ext at rank " << i <<std::endl;
            extDampers.push_back ( &attRobotCylinderExtDampers[i] );
        }

    for ( unsigned int i = 0;i<attLinkFootprintPlaneDampers.size();i++ )
        if ( checkLinkFootprintPlanePair ( i ) )
        {
            //std::cout << "active pair of LinkFootprint-Plane at rank " << i <<std::endl;
            linkDampers.push_back ( &attLinkFootprintPlaneDampers[i] );
        }

    attConstraint->matrix().resize ( dampers.size() +extDampers.size() + 2*linkDampers.size(),attCentipede->numberDof() );
    attConstraint->vector().resize ( dampers.size() +extDampers.size() + 2*linkDampers.size() );
    attConstraint->matrix().clear();
    attConstraint->vector().clear();

    for ( unsigned int i = 0;i<dampers.size();i++ )
        calcConstraint ( *dampers[i], i );

    for ( unsigned int i = 0;i<extDampers.size();i++ )
        calcExtConstraint ( *extDampers[i], dampers.size() +i );

    for ( unsigned int i = 0;i<linkDampers.size();i++ )
        calcLinkConstraint ( *linkDampers[i], dampers.size() +extDampers.size() + 2*i );
}



const vectorN& ChppLosCollisionAvoider::currentValue()
{
    attValue.clear();

    updateModel();

    std::vector<SelfDamper*> dampers;
    std::vector<ExtDamper*> extDampers;
    std::vector<LinkDamper*> linkDampers;
    for ( unsigned int i = 0;i<attRobotCylinderSphereDampers.size();i++ )
        if ( checkRobotCylinderSpherePair ( i ) )
    {
       // std::cout << "Cylinder-Sphere at rank " << i << std::endl;
            dampers.push_back ( &attRobotCylinderSphereDampers[i] );
    }

    for ( unsigned int i = 0;i<attRobotCylinderDampers.size();i++ )
        if ( checkRobotCylinderPair ( i ) )
    {
       // std::cout << "Cylinders at rank " << i << std::endl;
            dampers.push_back ( &attRobotCylinderDampers[i] );
    }

    for ( unsigned int i = 0;i<attRobotSphereDampers.size();i++ )
        if ( checkRobotSpherePair ( i ) )
    {
       // std::cout << "Spheres at rank " << i << std::endl;
            dampers.push_back ( &attRobotSphereDampers[i] );
    }

    for ( unsigned int i = 0;i<attRobotCylinderPlaneDampers.size();i++ )
        if ( checkRobotCylinderPlanePair ( i ) )
    {
        //std::cout << "Cylinder-Plane at rank " << i << std::endl;
            extDampers.push_back ( &attRobotCylinderPlaneDampers[i] );
    }

    for ( unsigned int i = 0;i<attRobotCylinderExtDampers.size();i++ )
        if ( checkRobotCylinderExtPair ( i ) )
    {
        //std::cout << "Cylinder-Ext at rank " << i << std::endl;
            extDampers.push_back ( &attRobotCylinderExtDampers[i] );
    }
    
    for ( unsigned int i = 0;i<attLinkFootprintPlaneDampers.size();i++ )
        if ( checkLinkFootprintPlanePair ( i ) )
    {
        //std::cout << "LinkFootprint - Plane at rank " << i << std::endl;
            linkDampers.push_back ( &attLinkFootprintPlaneDampers[i] );
    }


    attValue.resize ( dampers.size() +extDampers.size() + 2*linkDampers.size());
    for ( unsigned int i = 0;i<dampers.size();i++ )
        attValue ( i ) = ( dampers[i]->distance < dampers[i]->Ds ) ? ( dampers[i]->distance - dampers[i]->Ds ) :0;

    for ( unsigned int i = 0;i<extDampers.size();i++ )
        attValue ( dampers.size() +i ) = ( extDampers[i]->distance < extDampers[i]->Ds ) ? ( extDampers[i]->distance - extDampers[i]->Ds ) :0;
    
    for ( unsigned int i = 0;i<linkDampers.size();i++ )
    {
        attValue ( extDampers.size() + dampers.size() + 2*i ) = ( linkDampers[i]->distance1 < linkDampers[i]->Ds ) ? ( linkDampers[i]->distance1 - linkDampers[i]->Ds ) :0;
        
        attValue ( extDampers.size() + dampers.size() + 2*i +1) = ( linkDampers[i]->distance2 < linkDampers[i]->Ds ) ? ( linkDampers[i]->distance2 - linkDampers[i]->Ds ) :0;
    }

    //std::cout << attValue << std::endl;

    return attValue;

}



