/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpplocalstepper/tasks/hppLosPointingTask.hh"
#include "hppGik/hppGikTools.h"
using namespace boost::numeric::ublas;

#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

void ChppLosPointingTask::equivAsymMat3 ( const vector3d& inVec, matrixNxP& outMat )
{
    outMat ( 0,0 ) = 0;
    outMat ( 0,1 ) = -inVec ( 2 );
    outMat ( 0,2 ) = inVec ( 1 );
    outMat ( 1,0 ) = inVec ( 2 );
    outMat ( 1,1 ) = 0;
    outMat ( 1,2 ) = -inVec ( 0 );
    outMat ( 2,0 ) = -inVec ( 1 );
    outMat ( 2,1 ) = inVec ( 0 );
    outMat ( 2,2 ) = 0;
}

void ChppLosPointingTask::crossProd3 ( const vector3d& v1, const vector3d& v2, vector3d& retV )
{
    retV ( 0 ) = v1 ( 1 ) *v2 ( 2 ) - v1 ( 2 ) *v2 ( 1 );
    retV ( 1 ) = v1 ( 2 ) *v2 ( 0 ) - v1 ( 0 ) *v2 ( 2 );
    retV ( 2 ) = v1 ( 0 ) *v2 ( 1 ) - v1 ( 1 ) *v2 ( 0 );
}

void ChppLosPointingTask::Rodrigues ( const vector3d& inW, double inAngle, matrixNxP& outRot )
{
    double nor = MAL_S3_VECTOR_NORM ( inW );
    identity_matrix<double> id ( 3 );
    if ( nor < 1e-8 )
        outRot = id;
    else
    {
        vector3d wn = inW / nor;
        equivAsymMat3 ( wn,outRot );
        outRot = id + sin ( inAngle ) *outRot + ( 1-cos ( inAngle ) ) * prod ( outRot,outRot );
    }
}

ChppLosPointingTask::ChppLosPointingTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint,const vector3d& inLocalOrigin,const vector3d& inLocalVector,const vector3d& inWorlTarget ) :ChppLosTaskOnJoint ( inCentipede,inJoint )
{
    attRobotLink = &inRobotLink;
    attLocalOrigin = inLocalOrigin;
    attLocalVector = inLocalVector;
    attWorldTarget = inWorlTarget;

    attConstraint->matrix().resize ( 3,attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attConstraint->vector().resize ( 3 );
    attTempJacobian.resize ( 6, attCentipede->numberDof() );
    attValue.resize ( 1 );
    attMaxVelocity = M_PI/2;
}

bool ChppLosPointingTask::resize()
{
    if ( ChppLosTask::resize() )
    {
        attTempJacobian.resize ( 6,attCentipede->numberDof() );
        return true;
    }
    return false;
}

void ChppLosPointingTask::joint ( CjrlJoint& inJ )
{
    attJoint = &inJ;
}

const vector3d& ChppLosPointingTask::localOrigin()
{
    return attLocalOrigin;
}
void ChppLosPointingTask::localOrigin ( const vector3d& inVec )
{
    attLocalOrigin = inVec;
}
const vector3d& ChppLosPointingTask::localVector()
{
    return attLocalVector;
}
void ChppLosPointingTask::localVector ( const vector3d& inVec )
{
    attLocalVector = inVec;
}

const vector3d& ChppLosPointingTask::worldTarget()
{
    return attWorldTarget;
}

void ChppLosPointingTask::worldTarget ( const vector3d& inVec )
{
    attWorldTarget = inVec;
}

void ChppLosPointingTask::maxVelocityNorm ( double inMV )
{
    attMaxVelocity = ( inMV>0.0 ) ?inMV:0.0;
}

const vectorN& ChppLosPointingTask::currentValue()
{
    m =attJoint->currentTransformation();
    vector3d O = m*attLocalOrigin;
    matrix3d m3;
    for ( unsigned int i=0;i<3;i++ )
        for ( unsigned int j=0;j<3;j++ )
            M3_IJ ( m3 , i,j ) = M4_IJ ( m, i,j );

    vector3d OP = m3*attLocalVector;
    vector3d OT = attWorldTarget - O;

    double normOP = MAL_S3_VECTOR_NORM ( OP );
    double normOT = MAL_S3_VECTOR_NORM ( OT );

    if ( normOP != 0 )
        OP = OP/normOP;
    if ( normOT != 0 )
        OT = OT/normOT;

    vector3d cross;
    MAL_S3_VECTOR_CROSS_PRODUCT ( cross,OT,OP );
    double sv = MAL_S3_VECTOR_NORM ( cross );
    double cv = OP ( 0 ) *OT ( 0 ) + OP ( 1 ) *OT ( 1 ) + OP ( 2 ) *OT ( 2 );
    attValue ( 0 ) = atan2 ( sv,cv );

    return attValue;
}

void ChppLosPointingTask::computeConstraint()
{
    resize();
    //task vector
    m =attJoint->currentTransformation();
    vector3d O = m*attLocalOrigin;
    matrix3d m3;
    for ( unsigned int i=0;i<3;i++ )
        for ( unsigned int j=0;j<3;j++ )
            M3_IJ ( m3 , i,j ) = M4_IJ ( attJoint->currentTransformation(), i,j );
    vector3d OP = m3*attLocalVector;
    vector3d OT = attWorldTarget - O;
    double normOP = MAL_S3_VECTOR_NORM ( OP );
    double normOT = MAL_S3_VECTOR_NORM ( OT );

    vector3d cross;
    MAL_S3_VECTOR_CROSS_PRODUCT ( cross ,OT,OP );
    attConstraint->vector() ( 0 ) = cross ( 0 );
    attConstraint->vector() ( 1 ) = cross ( 1 );
    attConstraint->vector() ( 2 ) = cross ( 2 );
    double normcross = MAL_S3_VECTOR_NORM ( cross );

    double sv = normcross;
    double cv = OP ( 0 ) *OT ( 0 ) + OP ( 1 ) *OT ( 1 ) + OP ( 2 ) *OT ( 2 );
    if ( ( normOP != 0 ) && ( normOT != 0 ) )
    {
        double nor = normOP*normOT;
        sv /= nor;
        cv /= nor;
    }
    double angl = atan2 ( sv,cv );

    if ( angl > M_PI/2 - 1e-2 )
    {
        
        matrixNxP rot ( 3,3 );
        matrix3d rot3;
        Rodrigues ( cross, angl/2, rot );
        ChppGikTools::UblastoMatrix3 ( rot, rot3 );
        OT = rot3 * OT;
        MAL_S3_VECTOR_CROSS_PRODUCT ( cross ,OT,OP );
        attConstraint->vector() ( 0 ) = cross ( 0 );
        attConstraint->vector() ( 1 ) = cross ( 1 );
        attConstraint->vector() ( 2 ) = cross ( 2 );
        normcross = MAL_S3_VECTOR_NORM ( cross );
        
    }

    if ( normcross > attMaxVelocity )
        attConstraint->vector() = attMaxVelocity*attConstraint->vector() /normcross;


    //Body jacobian
    attTempJacobian.clear();
    attRobotLink->standingRobot().robot()->getJacobian ( *attRobotLink->rootJoint(),*attJoint, attLocalOrigin, attTempJacobian, attRobotLink->rank(), false );
    ChppLosLink* plink = attRobotLink->parent();
    m =attJoint->currentTransformation();
    attCurrentPoint = m * attLocalOrigin;
    while ( plink!=0 )
    {
        plink->getJacobian ( attCurrentPoint,attTempJacobian,plink->rank() );
        plink = plink->parent();
    }


    //task jacobian
    matrixNxP matOP ( 3,3 ), matOT ( 3,3 ),matOTOP ( 3,3 );
    equivAsymMat3 ( OT,matOT );
    equivAsymMat3 ( OP,matOP );
    noalias ( matOTOP ) = prod ( matOT,matOP );
    noalias ( attConstraint->matrix() ) = prod ( matOTOP,subrange ( attTempJacobian,3,6,0,attCentipede->numberDof() ) );
    noalias ( attConstraint->matrix() ) -= prod ( matOP,subrange ( attTempJacobian,0,3,0,attCentipede->numberDof() ) );
    /*
        subrange(attConstraint->matrix(),0,3,6,attCentipede->numberDof()) = zero_matrix<double>(3,attCentipede->numberDof()-6);
        */
    /*
    std::cout << subrange(attConstraint->matrix(),0,3,0,6) << std::endl;
    std::cout << subrange(attConstraint->vector(),0,3) << std::endl;
    */

}

